@@ -20,6 +20,10 @@
 #include <glib.h>
 #include <glib/gi18n.h>
 
+#ifdef HAS_URLDISPATCHER
+# include <url-dispatcher.h>
+#endif
+
 #include "dbus-end-session-dialog.h"
 #include "dbus-login1-manager.h"
 #include "dbus-webcredentials.h"
@@ -31,6 +35,11 @@
 
 #include "../utils.h"
 
+/* some prototypes... */
+static char *   find_browser ();
+static void     run_outside_app (const char * cmd);
+static gboolean zenity_warning (const char * icon_name, const char * title, const char * text);
+
 enum
 {
   END_SESSION_TYPE_LOGOUT = 0,
@@ -94,10 +103,24 @@ typedef enum
   PROMPT_WITH_ZENITY,
   PROMPT_WITH_AYATANA,
   PROMPT_WITH_MATE,
+  PROMPT_WITH_XFCE,
 }
 prompt_status_t;
 
 
+static gboolean
+have_gnome_program (const gchar *program)
+{
+  g_auto(GStrv) desktop_names = NULL;
+
+  if (is_gnome()) {
+    g_autofree gchar *path = g_find_program_in_path (program);
+    return path != NULL;
+  }
+
+  return FALSE;
+}
+
 static gboolean
 have_mate_program (const gchar *program)
 {
@@ -111,6 +134,19 @@ have_mate_program (const gchar *program)
   return FALSE;
 }
 
+static gboolean
+have_xfce_program (const gchar *program)
+{
+  g_auto(GStrv) desktop_names = NULL;
+
+  if (is_xfce()) {
+    g_autofree gchar *path = g_find_program_in_path (program);
+    return path != NULL;
+  }
+
+  return FALSE;
+}
+
 static prompt_status_t
 get_prompt_status (IndicatorSessionActionsDbus * self)
 {
@@ -123,6 +159,10 @@ get_prompt_status (IndicatorSessionActionsDbus * self)
       if ((prompt == PROMPT_NONE) && have_mate_program ("mate-session-save"))
           prompt = PROMPT_WITH_MATE;
 
+      /* can we use the XFCE prompt? */
+      if ((prompt == PROMPT_NONE) && have_xfce_program ("xfce4-session-logout"))
+          prompt = PROMPT_WITH_XFCE;
+
       /* can we use the Unity/Ayatana prompt? */
       if ((prompt == PROMPT_NONE) && p && p->end_session_dialog)
         {
@@ -737,6 +777,33 @@ zenity_question (IndicatorSessionActionsDbus * self,
   return confirmed;
 }
 
+static void
+my_bug (IndicatorSessionActions * self G_GNUC_UNUSED)
+{
+  const  char * bts_url = get_distro_bts_url();
+  static char * browser = NULL;
+
+#ifdef HAS_URLDISPATCHER
+  if (g_getenv ("MIR_SOCKET") != NULL)
+    url_dispatch_send(bts_url, NULL, NULL);
+  else
+#endif
+    {
+
+      if (browser == NULL)
+        browser = find_browser();
+
+      if (browser != NULL)
+        run_outside_app(g_strdup_printf("%s '%s'", browser, bts_url));
+
+      else
+        zenity_warning ("dialog-warning",
+                        _("Warning"),
+                        _("The operating system's bug tracker needs to be accessed with\na web browser.\n\nThe Ayatana Session Indicator could not find any web\nbrowser on your computer."));
+
+    }
+}
+
 static gboolean
 zenity_warning (const char * icon_name,
                 const char * title,
@@ -809,6 +876,10 @@ my_logout (IndicatorSessionActions * actions)
         run_outside_app ("mate-session-save --logout-dialog");
         break;
 
+      case PROMPT_WITH_XFCE:
+        run_outside_app ("xfce4-session-logout");
+        break;
+
       case PROMPT_NONE:
         logout_now (self);
         break;
@@ -851,6 +922,10 @@ my_reboot (IndicatorSessionActions * actions)
         run_outside_app ("mate-session-save --shutdown-dialog");
         break;
 
+      case PROMPT_WITH_XFCE:
+        run_outside_app ("xfce4-session-logout");
+        break;
+
       case PROMPT_NONE:
         reboot_now (self);
         break;
@@ -888,6 +963,10 @@ my_power_off (IndicatorSessionActions * actions)
         run_outside_app ("mate-session-save --shutdown-dialog");
         break;
 
+      case PROMPT_WITH_XFCE:
+        run_outside_app ("xfce4-session-logout");
+        break;
+
       case PROMPT_WITH_ZENITY:
         if (zenity_question (self,
               "system-shutdown",
@@ -908,15 +987,6 @@ my_power_off (IndicatorSessionActions * actions)
 ****
 ***/
 
-static void
-my_desktop_help (IndicatorSessionActions * self G_GNUC_UNUSED)
-{
-  if (have_mate_program ("yelp"))
-    run_outside_app ("yelp help:mate-user-guide");
-  else
-    run_outside_app ("yelp");
-}
-
 static char *
 find_browser ()
 {
@@ -948,6 +1018,32 @@ find_browser ()
 
 }
 
+static void
+my_desktop_help (IndicatorSessionActions * self G_GNUC_UNUSED)
+{
+  static char * browser = NULL;
+
+  if (have_gnome_program ("yelp"))
+    run_outside_app ("yelp help:gnome-user-guide");
+  else if (have_mate_program ("yelp"))
+    run_outside_app ("yelp help:mate-user-guide");
+  else if (is_xfce())
+    {
+      if (browser == NULL)
+        browser = find_browser();
+      if (browser != NULL)
+        run_outside_app(g_strdup_printf("%s '%s'", browser, "https://docs.xfce.org/"));
+      else
+        zenity_warning ("dialog-warning",
+                        _("Warning"),
+                        _("The XFCE desktop's user guide needs to be accessed with\na web browser.\n\nThe Ayatana Session Indicator could not find any web\nbrowser on your computer."));
+    }
+  else
+    zenity_warning ("dialog-warning",
+                    _("Warning"),
+                    _("The Ayatana Session Indicator does not know yet, how to show\nthe currently running desktop's user guide or help center.\n\nPlease report this to the developers at:\nhttps://github.com/ArcticaProject/ayatana-indicator-session/issues"));
+}
+
 static void
 my_distro_help (IndicatorSessionActions * self G_GNUC_UNUSED)
 {
@@ -958,6 +1054,10 @@ my_distro_help (IndicatorSessionActions * self G_GNUC_UNUSED)
 
   if (browser != NULL)
     run_outside_app(g_strdup_printf("%s '%s'", browser, get_distro_url()));
+  else
+    zenity_warning ("dialog-warning",
+                    _("Warning"),
+                    g_strdup_printf(_("Displaying information on %s  requires\na web browser.\n\nThe Ayatana Session Indicator could not find any web\nbrowser on your computer."), get_distro_name()));
 }
 
 
@@ -996,12 +1096,19 @@ have_gnome_control_center (void)
 static void
 my_settings (IndicatorSessionActions * self G_GNUC_UNUSED)
 {
+#ifdef HAS_URLDISPATCHER
+  if (g_getenv ("MIR_SOCKET") != NULL)
+    url_dispatch_send("settings:///system", NULL, NULL);
+  else
+#endif
   if (have_unity_control_center ())
     run_outside_app ("unity-control-center");
   else if (have_gnome_control_center())
     run_outside_app ("gnome-control-center");
   else if (have_mate_program ("mate-control-center"))
     run_outside_app ("mate-control-center");
+  else if (have_xfce_program ("xfce4-settings-manager"))
+    run_outside_app ("xfce4-settings-manager");
   else
     zenity_warning ("dialog-warning",
                     _("Warning"),
@@ -1011,6 +1118,11 @@ my_settings (IndicatorSessionActions * self G_GNUC_UNUSED)
 static void
 my_online_accounts (IndicatorSessionActions * self G_GNUC_UNUSED)
 {
+#ifdef HAS_URLDISPATCHER
+  if (g_getenv ("MIR_SOCKET") != NULL)
+    url_dispatch_send("settings:///system/online-accounts", NULL, NULL);
+  else
+#endif
   if (have_unity_control_center ())
     run_outside_app ("unity-control-center credentials");
   else if (have_gnome_control_center())
@@ -1024,12 +1136,19 @@ my_online_accounts (IndicatorSessionActions * self G_GNUC_UNUSED)
 static void
 my_about (IndicatorSessionActions * self G_GNUC_UNUSED)
 {
+#ifdef HAS_URLDISPATCHER
+  if (g_getenv ("MIR_SOCKET") != NULL)
+    url_dispatch_send("settings:///system/about", NULL, NULL);
+  else
+#endif
   if (have_unity_control_center ())
     run_outside_app ("unity-control-center info");
   else if (have_gnome_control_center())
     run_outside_app ("gnome-control-center info");
   else if (have_mate_program ("mate-system-monitor"))
     run_outside_app ("mate-system-monitor --show-system-tab");
+  else if (have_xfce_program ("xfce4-about"))
+    run_outside_app ("xfce4-about");
   else
     zenity_warning ("dialog-warning",
                     _("Warning"),
@@ -1060,6 +1179,10 @@ lock_current_session (IndicatorSessionActions * self, gboolean immediate)
     {
       run_outside_app ("mate-screensaver-command --lock");
     }
+  else if (have_xfce_program ("xflock4"))
+    {
+      run_outside_app ("xflock4");
+    }
   else
     {
       g_return_if_fail (p->screen_saver != NULL);
@@ -1205,6 +1328,7 @@ indicator_session_actions_dbus_class_init (IndicatorSessionActionsDbusClass * kl
   actions_class->online_accounts = my_online_accounts;
   actions_class->desktop_help = my_desktop_help;
   actions_class->distro_help = my_distro_help;
+  actions_class->bug = my_bug;
   actions_class->about = my_about;
   actions_class->switch_to_screensaver = my_switch_to_screensaver;
   actions_class->switch_to_greeter = my_switch_to_greeter;
